package de.mlessmann.certassist.service;

import de.mlessmann.certassist.except.CommandLineOperationException;
import de.mlessmann.certassist.models.*;
import de.mlessmann.certassist.openssl.*;
import de.mlessmann.certassist.repositories.CertificateRepository;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import lombok.RequiredArgsConstructor;
import org.apache.commons.lang3.StringUtils;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.stereotype.Service;

@Service
@RequiredArgsConstructor
public class CertificateCreationService {

    private final CertificateRepository certificateRepository;
    private final OpenSSLService openSSLService;
    private final PassphraseService passphraseService;

    public Certificate createCertificate(final CertificateInfo certificateInfo) {
        final Certificate certificate = createEntityFromRequest(certificateInfo);

        try (OpenSSLCertificateResult certificateCreatorResult = openSSLService.createCertificate(certificateInfo);) {
            Path keyPath = certificateCreatorResult.certificateKeyPath();
            if (keyPath != null) {
                certificate.setPrivateKey(Files.readAllBytes(keyPath));
            }
            certificate.setCert(Files.readAllBytes(certificateCreatorResult.certificatePath()));
        } catch (CommandLineOperationException | IOException e) {
            throw new IllegalStateException("Failed to create certificate!", e);
        }

        certificateRepository.save(certificate);
        return certificate;
    }

    private Certificate createEntityFromRequest(CertificateInfo certificateInfo) {
        final Certificate certificate = new Certificate();
        certificate.setType(mapCertificateRequestType(certificateInfo.getType()));
        certificate.setSubjectCommonName(certificateInfo.getSubject().getCommonName());
        certificate.setTrustingAuthority(certificateInfo.getTrustingAuthority());
        certificate.setRequestedKeyLength(certificateInfo.getRequestedKeyLength());
        certificate.setRequestedValidityDays(certificateInfo.getRequestedValidityDays());
        final CertificateInfoSubject subjectInfo = certificateInfo.getSubject();
        certificate.setSubjectEmailAddress(subjectInfo.getEmailAddress());
        certificate.setSubjectOrganization(subjectInfo.getOrganization());
        certificate.setSubjectOrganizationalUnit(subjectInfo.getOrganizationalUnit());
        certificate.setSubjectCountry(subjectInfo.getCountry());
        certificate.setSubjectState(subjectInfo.getState());
        certificate.setSubjectLocality(subjectInfo.getLocality());

        final CertificateInfoExtension extension = certificateInfo.getExtension();
        if (extension != null) {
            final CertificateExtension certificateExtension = new CertificateExtension();
            certificateExtension.setIdentifier("alternativeNames");
            certificateExtension.setValue(String.join(",", extension.getAlternativeDnsNames()));
            certificate.setCertificateExtension(List.of(certificateExtension));
        }
        return certificate;
    }

    @NonNull
    public Certificate importCertificate(
        @NonNull Path certificate,
        @Nullable Path keyFile,
        @Nullable String keyPassphrase
    ) {
        try {
            String fingerprint = openSSLService.getCertificateFingerprint(certificate);
            var generatedRequest = openSSLService.getCertificateInfo(certificate);
            Certificate entity = createEntityFromRequest(generatedRequest);
            entity.setCert(Files.readAllBytes(certificate));
            if (keyFile != null) {
                entity.setPrivateKey(Files.readAllBytes(keyFile));
            }
            if (StringUtils.isNotBlank(keyPassphrase)) {
                passphraseService.storePassphrase("cert:" + fingerprint, keyPassphrase);
            }
            return certificateRepository.save(entity);
        } catch (CommandLineOperationException | IOException e) {
            throw new RuntimeException("Unable to import certificate", e);
        }
    }

    private CertificateType mapCertificateRequestType(CertificateInfo.RequestType requestType) {
        return switch (requestType) {
            case ROOT_AUTHORITY -> CertificateType.ROOT_CA;
            case STANDALONE_CERTIFICATE -> CertificateType.STANDALONE_CERT;
            case NORMAL_CERTIFICATE -> CertificateType.SIGNED_CERT;
        };
    }
}
